#!/usr/bin/env python

# test_where_am_i.py
# Copyright 2008 Shane Harper (http://shaneharper.net)
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#
# TDD (Test Driven Development) is being used to develop "where_am_i".

from where_am_i import where_am_i, indent_column
import StringIO
import sys
import unittest


class WhereAmI_TestCaseBase(unittest.TestCase):
    def check(self, python_code, line_number, expected_output):
        self.assertEqual(where_am_i(StringIO.StringIO(python_code),
                                    line_number), expected_output)


###############################################################################

class BasicTests(WhereAmI_TestCaseBase):
    """Tests without multi-line logical lines."""

    program1 = """# line 1
class MyClass:
    def __init__(self):
        print "Here"
    def fn(self):
        return 42
class MyClass2:
    def __init__(self):
        pass

if __name__ == '__main__':

    m = MyClass()
# comment
    print

    for i in range(42):
        while True:
            pass
"""

    def test_first_line(self):
        self.check(self.program1, 1, '')

    def test_comment_is_not_a_statement(self):
        self.check(self.program1, 2, '')

    def test_inside_class(self):
        self.check(self.program1, 3, 'class MyClass:\n')

    def test_inside_function(self):
        self.check(self.program1, 6, 'class MyClass:\n    def fn(self):\n')

    def test_function_and_class_end_on_same_line(self):
        self.check(self.program1, 7, '')

    def test_function_and_class_end_on_same_line_2(self):
        self.check(self.program1, 9,
                   'class MyClass2:\n    def __init__(self):\n')

    def test_function_and_class_end_on_same_line_3(self):
        self.check(self.program1, 12, "if __name__ == '__main__':\n")

    def test_blank_line_does_not_finish_statement(self):
        self.check(self.program1, 13, "if __name__ == '__main__':\n")

    def test_only_classes_functions_and_conditions_output(self):
        # Initially I expect "m = MyClass()" to be incorrectly output.
        self.check(self.program1, 14, "if __name__ == '__main__':\n")

    def test_for_and_while(self):
        self.check(self.program1, 19, "if __name__ == '__main__':\n"
                                      "    for i in range(42):\n"
                                      "        while True:\n")

    def test_single_line_class(self):
        # This test will initially fail: Previously a class only ended when a
        # line was encountered that was indented the same as (or less than) the
        # first line of the class.
        self.check('class SingleLineClass: pass\n\n', 2, '')

    def test_single_line_function(self):
        # This test is similar to test_single_line_class.
        self.check('class B:\n def do(self): throw NotImplementedError\n\n', 3,
                   'class B:\n')

    # XXX self.check('class MyClass: \\\npass\n',2, 'class MyClass:\n')


###############################################################################

class IndentationTests(WhereAmI_TestCaseBase):
    def test_tab(self):
        self.check("if answer == 42:\n"
                   "    if question_is_known:\n"
                   "\tprint 'Hooray'  # 8 spaces indentation due to tab!\n",
                   3,
                   "if answer == 42:\n    if question_is_known:\n")

    def test_tab2(self):
        # A tab moves to the next column which is divisible by 8.
        # (A tab is not always replaced by eight spaces, e.g. if it's proceeded
        # by other characters.)
        self.assertEqual(indent_column("  \tif a==b:"), 8)

# XXX def test_invalid_indentation(self):
#         self.check("""class MyClass:
#         def __init__(self): pass
#       def fn(self):
#           pass
#     """, 4,
#         "Invalid indentation encountered on line 3.")

    def test_form_feed(self):
        self.assertEqual(indent_column("\fif true:"), 0)

    def test_form_feed_proceeded_by_other_white_space(self):
        self.assertEqual(indent_column("\t \f  if true:"), 2)


###############################################################################

class MultipleLineLogicalLineTests(WhereAmI_TestCaseBase):
    def test_line_continuation_character(self):
        self.check("if a == \\\nb:\n  pass\n", 3,
                   "if a == \\\nb:\n")

    def test_line_continuation_character_on_last_line(self):
        # Check that a line continuation char at the end of the last line
        # doesn't cause problems. (When this test was first run the test went
        # into an endless loop!)
        self.check("if \\\n", 2, "if \\\n")

    def test_ignore_line_continuation_char_in_comment(self):
        self.check("def fn(): # single logical line \\\n  pass\n", 2,
                   "def fn(): # single logical line \\\n")

    def test_hash_inside_a_string(self):
        # The hash symbol on line 1 must not be interpreted as starting a
        # comment, otherwise the line continuation character ('\') will be
        # ignored.
        self.check('if a == "#" \\\n  and n == 42:\n  pass\n', 3,
                   'if a == "#" \\\n  and n == 42:\n')

    def test_escaped_quote_in_string(self):
        # Similar to test_hash_inside_a_string.
        self.check('if a == "\\"#" \\\n  and n == 42:\n  pass\n', 3,
                   'if a == "\\"#" \\\n  and n == 42:\n')

    def test_multi_line_string_with_continuation_character(self):
        # (This test passed initially: A '\' at the end of a line was handled
        # the same way regardless of whether it's inside a string or not.)
        self.check('if a == "abc\\\n123":\n  pass\n', 3,
                   'if a == "abc\\\n123":\n')

    def test_hash_inside_multi_line_string(self):
        self.check('if a == "abc\\\n:#\\\n123":\n  pass\n', 4,
                   'if a == "abc\\\n:#\\\n123":\n')

    def test_multi_line_triple_double_quote_string(self):
        self.check('if a == """\n#:\n""":\n  pass\n', 4,
                   'if a == """\n#:\n""":\n')

    def test_multi_line_triple_single_quote_string(self):
        self.check("if a == '''\n#:\n''':\n  pass\n", 4,
                   "if a == '''\n#:\n''':\n")

    def test_indentation_on_continuation_lines_ignored(self):
        # The indentation of line 3 must not cause the 'if' to end.
        self.check("if True:\n  a = '''\nabc123\n'''\n  print", 5,
                   "if True:\n")

    def test_multi_line_raw_strings(self):
        # Note: there's no special handling required for raw strings because
        # raw strings can't end with a '\'.
        self.check('if a == r"""\nr\\"""\n""":\n  pass\n', 4,
                   'if a == r"""\nr\\"""\n""":\n')

    program = ("def fn(\n"
               "    a, b\n"
               ")\\\n"
               ":\n"
               "    if a == \\\n"
               "        b:\n")

    def test_bracket_implicit_continuation_line(self):
        self.check(self.program+"        pass\n", 7, self.program)

    def test_implicit_continuation_line_ignore_bracket_within_strings(self):
        self.check('def fn(s = "):"\n):\n pass\n', 3,
                   'def fn(s = "):"\n):\n')

    def test_square_bracket_implicit_continuation_line(self):
        self.check('if a == [\n1, 42,\n]:\n  pass\n', 4,
                   'if a == [\n1, 42,\n]:\n')

    def test_braces_implicit_continuation_line(self):
        self.check('if a == {\n1: 42,\n}:\n  pass\n', 4,
                   'if a == {\n1: 42,\n}:\n')

    def test_braces_implicit_continuation_line_with_comment(self):
        self.check('if a == { #comment\n1: 42,\n}:\n  pass\n', 4,
                   'if a == { #comment\n1: 42,\n}:\n')

    def test_nested_brackets_implicit_continuation_line(self):
        self.check('def fn(a = {0:[\n1,2], b="}])"}\n):\n  pass', 4,
                   'def fn(a = {0:[\n1,2], b="}])"}\n):\n')

    def test_stop_in_the_middle_of_a_logical_line(self):
        # This test initially failed: where_am_i would go through to the end of
        # the source file if the stop line wasn't at the end of a logical line.
        self.check('def fn():\n  a = \\\n42\n\ndef fn2(): pass\n', 2,
                   'def fn():\n')

    def test_ignore_python_keywords_within_strings(self):
        # This test passed initially.
        # 'where_am_i' could have searched backwards from the stop-line looking
        # for lines that start with keywords such as 'if' or 'class' but could
        # have been tricked by keywords within strings.
        self.check('a = """\nif the weather is nice today:\n"""\n\n', 4, '')

    def test_ignore_python_keywords_within_strings_2(self):
        self.check('a == """\nclass Klass:\n"""\n\n', 4, '')


###############################################################################

class ComplexConditionsTests(WhereAmI_TestCaseBase):
    """Tests for conditions that include else, elif, etc."""

    program =  \
"""class MyClass:
    def __init__(self):
        print "Here"
        if n == 42:
            do_it()
        else:
            do_something()

n = 42
"""

    def test_sanity_check(self):
        self.check(self.program, 4, "class MyClass:\n    def __init__(self):\n")

    def test_inside_if_else(self):
        # when within an "else" corresponding "if" is output
        self.check(self.program, 7, "class MyClass:\n"
                                    "    def __init__(self):\n"
                                    "        if n == 42:\n"
                                    "        else:\n")

    def test_else_finished(self):
        self.check(self.program, 9, '')

    program2 = ("if a:\n"
                "    print 'a'\n"
                "elif b:\n"
                "    print 'b'\n"
                "elif c:\n"
                "    print 'c'\n"
                "else:\n"
                "    print\n\n")

    def test_elif_followed_by_another_elif(self):
        # Both "elif"s are output because they both can influence whether the
        # statement after the second "elif" is executed (because conditions b
        # and c need not be mutually exclusive.)
        # In program2, 'print c' is executed only if b is false and c is true
        # (both conditions b and c are relevant).
        self.check(self.program2, 6, 'if a:\nelif b:\nelif c:\n')

    def test_else_following_elif(self):
        self.check(self.program2, 8, 'if a:\nelif b:\nelif c:\nelse:\n')

    def test_else_following_for(self):
        self.check('for i in range(42):\n  pass\nelse:\n  print\n', 4,
                   'for i in range(42):\nelse:\n')

    def test_else_following_while(self):
        self.check('while get_result():\n  pass\nelse:\n  print\n', 4,
                   'while get_result():\nelse:\n')

    def test_statement_following_if_on_one_line(self):
        # (This test is related to test_single_line_class and
        # test_single_line_function.)
        self.check('if a: n = 42\nelse:\n  n = 6*7;\n', 3,
                   'if a: n = 42\nelse:\n')
        #          'if a:\nelse:\n') XXX ?  "n = 42" is not a condition!

    def test_try_except(self):
        self.check("def my_fn():\n"
                   "    try:\n"
                   "        do_it()\n"
                   "    except Disaster:\n"
                   "        print 'Oh, my!'\n",
                   5,
                   "def my_fn():\n"
                   "    try:\n"
                   "    except Disaster:\n")

    def test_except_following_except(self):
        # "except"s at the end of a try tend to be mutually exclusive (the
        # first except that an exception matches is likely to be the only one),
        # so there's little point in listing the proceeding "except"s.
        self.check('''def my_fn():
    try:
        do_it()
    except Disaster:
        print 'Yikes!'
    except Catastrophe:
        print 'Kaboom!'
''', 7, '''def my_fn():
    try:
    except Catastrophe:
''')

    def test_else_following_try(self):
        # There is no need to output the "except" statement when inside the
        # "else" because the exceptions that can be caught have no influence on
        # whether the "else" statement is executed or not.
        self.check('''def my_fn():
    try:
        do_it()
    except Disaster:
        print 'Oh, my!'
    else:
        print 'Ok!'
''', 7, '''def my_fn():
    try:
    else:
''')

    def test_finally_statement(self):
        # The "finally" code is unconditionally reached if my_fn is called...
        self.check('''def my_fn():
    try:
        do_it()
    except Disaster:
        print 'Oh, my!'
    finally:
        print 'finally'\n''', 7, 'def my_fn():\n')


###############################################################################

class EarlyTerminationOfCompoundStatementTests(WhereAmI_TestCaseBase):
    def test_return_statement(self):
        # The conditions leading up to a "return" specify (at least in part)
        # conditions for which the following statements of the function can be
        # reached.
        self.check("def my_fn():\n if a:\n  return\n  print\n", 4,
                   "def my_fn():\n if a:\n  return\n")

    def test_return_statement_2(self):
        # Conditions leading up to a "return" in one function are not relevant
        # in another function.
        self.check("def my_fn():\n if a:\n  return\n  print\n"
                   "def my_fn2(): pass\n", 5, '')

    def test_return_on_same_line_as_if(self):
        self.check('def my_fn():\n  if a: return\n  print "Here"\n', 3,
                   'def my_fn():\n  if a: return\n')

# XXX def test_return_on_same_line_as_if_2(self):
#         self.check('def my_fn():\n if a: print "a"; return\n print\n', 3,
#                    'def my_fn():\n if a: print "a"; return\n')
# XXX Should print statement be output?

    def break_and_continue_tests(self, loop_stmt, break_or_continue_stmt):
        # "break"/"continue" within loop
        self.check(
                loop_stmt+'\n if c:\n  '+break_or_continue_stmt+'\n print\n', 4,
                loop_stmt+'\n if c:\n  '+break_or_continue_stmt+'\n')

        # "break"/"continue" on a line with another statement within loop
        self.check(
                loop_stmt+'\n if c: '+break_or_continue_stmt+'\n print\n', 3,
                loop_stmt+'\n if c: '+break_or_continue_stmt+'\n')

        # conditions leading up to a "break"/"continue" are not relevant outside of
        # the loop
        self.check(
                loop_stmt+'\n if c: '+break_or_continue_stmt+'\n print\nn=42\n',
                4, '')

    def test_break_in_for_loop(self):
        self.break_and_continue_tests('for i in range(42):', 'break')

    def test_continue_in_for_loop(self):
        self.break_and_continue_tests('for i in range(42):', 'continue')

    def test_break_in_while_loop(self):
        self.break_and_continue_tests('while 1:', 'break')

    def test_continue_in_while_loop(self):
        self.break_and_continue_tests('while 1:', 'continue')

    def test_break_within_nested_loops(self):
        self.check('for a in b:\n while 1:\n  if c: break\n print\n', 4,
                   'for a in b:\n')

    def test_return_and_break_within_a_loop(self):
        # The conditions leading up to the "return" are output, as is the
        # "break" within the loop because should the loop be exited due to
        # the "break" then this could preclude the "return" being executed.
        self.check('''def my_fn():
    for i in l:
        if b(): return
        if c(): break
    if e:
        x = y
''', 6, '''def my_fn():
    for i in l:
        if b(): return
        if c(): break
    if e:
''')

    def test_raise(self):
        self.check("def my_fn():\n if a:\n  raise MyException\n print\n", 4,
                   "def my_fn():\n if a:\n  raise MyException\n")

    def test_raise_statement_outside_of_a_function(self):
        self.check("if a:\n  raise MyException\nprint\n", 3,
                   "if a:\n  raise MyException\n")

# XXX def test_except_raised_exception(self):
#        self.check("""def my_fn():
#            try: 
#                raise MyException()
#            except: pass
#            print""", 5, "def my_fn()")

# XXX def test_yield(self):
#         self.check('''def my_generator():
#         for v in x:
#             yield v
#             do_it()
#     ''', 4,
#     '''def my_generator():
#         for v in x:
#             yield v
#     ''')

# XXX Presently assert statements aren't output by where_am_i: They are
# XXX generally not interesting because normally asserts should never fail.
#
# XXX def test_assert(self):
#         self.check('def my_fn():\n if answer:\n assert answer == 42\n print', 4,
#                    'def my_fn():\n if answer:\n assert answer == 42\n')


if __name__ == '__main__':
    unittest.main()
